import scipy as sp
import numpy as np
from sklearn.metrics import mean_absolute_error, mean_squared_error, roc_auc_score
from bliz.evaluation.constants import MetricsConstants
from bliz.evaluation.utils import close_enough,\
    validate_multiple_lists_length,\
    samples_to_bin_numbers,\
    is_binary

MSE = mean_squared_error
ABS_ERR = mean_absolute_error
ERR_TYPES = ["mse", "abs"]
WD = sp.stats.wasserstein_distance


def discriminability(
        a,
        mean,
        rand,
):
    """calculates the discrimination of model given error distribution a,
    against random and mean `models`

    Args:
        a (array-like): the error distribution of model
        mean (array-like): the error distribution against `guessing` the mean value
        rand (array-like): the error distribution against `guessing` random values

    Returns:
        d(float): the rate of discrimination

    """
    if np.mean(a) > np.mean(mean):
        return 0.0
    else:
        MX = WD(np.asarray([0.0] * len(a)), mean)
        B = WD(a, mean)
        C = WD(mean, rand)

        return (B / C) / (MX / C)


def certainty(
        a,
        rand,
):
    """the inverse of the standard deviation of the error distribution of a model

    Args:
        a (array-like): the error distribution of model
        rand (array-like): the error distribution against `guessing` random values

    Returns:
        c (float): the std^-1
    """
    std_a = np.std(a)
    std_rand = np.std(rand)
    if close_enough(std_a, 0.0, 16):
        return -1
    else:
        return float(std_rand / std_a)


def divergency(s):
    """the average distribution distance between the ground truth and the predictions
    veotrs

    Args:
        s (array-like): vector of the distances of each sample

    Returns:
        d (float): mean value of distnace
    """
    return float(np.mean(s))


def _weighted_error(
        bins,
        weights: np.ndarray,
        how: str,
):
    """returns an error calculating function according to weights of samples
    given by the weights matrix

    Args:
        bins (array-like): the bin edges for grouping the r, p vectors
        weights (np.ndarray): an N X N matrix representing the cost of predicting
                              in different bin
        how (str): should the error be absolute / squared, gets values in ['abs', 'mse']
                   defaults to 'mse'

    Returns:
        _calc_metric (callable): a functions that gets two vectors and returns the calculated error
                                 with accordance to bins, weights and how
    """
    if weights.shape == (len(bins) - 1, len(bins) - 1):
        _step = (lambda i, j: abs(i - j)) if how == 'abs' else (
            lambda i, j: ((i - j)**2))

        def _calc_metric(r, p):
            if validate_multiple_lists_length(r, p):
                weighted_sum = 0.0
                sum_of_weights = 0.0
                r_bins, p_bins = samples_to_bin_numbers(r, p, bins=bins)
                for i, j, ib, jb in zip(r, p, r_bins, p_bins):
                    w = weights[ib, jb]
                    diff = _step(i, j)
                    weighted_sum += diff * w
                    sum_of_weights += w
                return (weighted_sum / sum_of_weights)
            else:
                raise ValueError(
                    "r, p should be of same length but len(r) = {lr}, len(p) = {lp}"
                    .format(lr=len(r), lp=len(p)))

        return _calc_metric
    else:
        raise IndexError(
            "size of weights matrix should be same as len(bins) - 1: {} X {}".
            format(len(bins) - 1,
                   len(bins) - 1))


def _resize_matrix_function(mx: np.ndarray, kind='linear'):
    """creates an interpolation function with accordance to the matrix mx

    Args:
        mx (np.ndarray): a cubic matrix
        kind (str): type of interpolation, gets values in linear, cubic

    Returns:
        f (callable): an interpolation function
    """
    x = np.linspace(0, 1, mx.shape[0])
    y = np.linspace(0, 1, mx.shape[1])
    f = sp.interpolate.interp2d(x, y, mx, kind=kind)
    return f


def _resize_vector_function(v):
    """creates an interpolation function with accordance to the vector v.
    with which we could expand the vector v

    Args:
        v (array-like): a 1d vector

    Returns:
        f (callable): an interpolation function in 1d
    """
    _v = np.asarray(v)
    f = sp.interpolate.interp1d(
        np.linspace(_v.min(), _v.max(), len(v)), v, kind='cubic')
    return f


def _interpolate_weights(
        w: np.ndarray,
        new_size: int,
):
    """extends a cubic weight matrix to new size

    Args:
        w (np.ndarray): a cubic matrix of weights
        new_size (int): the new size of output matrix

    Returns:
        _w (np.ndarray): a cubic weight matrix with size new_size
    """
    if new_size == w.shape[0]:
        return w
    else:
        _w = w
        if max(w.shape) <= 3:
            _w = _resize_matrix_function(w)(np.linspace(0, 1, 4),
                                            np.linspace(0, 1, 4))
        _w = _resize_matrix_function(
            _w, kind='cubic')(np.linspace(0, 1, new_size),
                              np.linspace(0, 1, new_size))
        for i in range(new_size):
            _w[i, i] = 1
    return _w


def _interpolate_bins(bins, new_size: int):
    """extends a bins vector

    Args:
        bins (array-like): bin edges vector
        new_size (int): the new size of output matrix

    Returns:
        new_bins (array-like): a cubic weight matrix with size new_size
    """
    bins_as_vec = np.asarray(bins)
    new_bins_x = np.linspace(bins_as_vec.min(), bins_as_vec.max(), new_size)
    new_bins = _resize_vector_function(bins)(new_bins_x)
    return new_bins


def get_weight_metric(bins, weights: np.ndarray,
                      new_size) -> (callable, np.ndarray):
    """returns an interpolated weight metric

    Args:
        bins(array-like): array of bin edges
        weights(np.ndarray): a small (3X3 usually) matrix of bin-weight errors
        new_size(int): interpolation new size ( > len(bins)-1)

    Returns:
        T(tuple[callable, np.ndarray]): a tuple where the first arg is a function to calculate
            weights between two scores and the second is matrix of new weights
    """
    new_bins = _interpolate_bins(bins, new_size + 1)
    new_weights = _interpolate_weights(weights, new_size)

    def _sample_weight(a, b):
        a_bins, b_bins = samples_to_bin_numbers(
            [a], [b], bins=new_bins
        )  # TODO: change samples to bins to handle list and scalar
        return new_weights[a_bins[0], b_bins[0]]

    return _sample_weight, new_weights


def weighted_interpolated_error(
        size,
        bins,
        weights: np.ndarray,
        error_type: str,
):
    """returns a error calculation function

    Args:
        r_vector(array-like): result vector
        bins(list): bin edges vector
        weights(np.ndarray): matrix of weights between bins
        error_type(str): type of error to use, (abs/mse)

    Returns:
        _calc_metric (callable): a functions that gets two vectors and returns the calculated error
                                 with accordance to bins, weights and how
    """
    if error_type in ERR_TYPES:
        new_size = max(9, size)
        new_bins = _interpolate_bins(bins, new_size + 1)
        new_weights = _interpolate_weights(weights, new_size)
        return _weighted_error(new_bins, new_weights, how=error_type)
    else:
        raise ValueError("error type must be in {}".format(ERR_TYPES))


def inverse_accuracy(y_real, y_pred):
    def inverse_acc(y_r, y_p):
        ineq_sum = 0
        for i, j in zip(y_r, y_p):
            if i != j:
                ineq_sum += 1
        return ineq_sum / len(y_real)

    if len(y_pred) == len(y_real):
        if is_binary(y_pred):
            return inverse_acc(y_real, y_pred)
        else:
            if (len(y_pred.shape) == 2 and y_pred.shape[1] == 2):
                return inverse_acc(y_real, np.argmax(y_pred, axis=1))
            else:
                return inverse_acc(y_real, np.round(y_pred))


def inverse_roc_auc(y_real, y_pred):
    if is_binary(y_pred):
        raise ValueError("y_pred should be (n, 2) shaped probability vector")
    if len(y_pred.shape) == 2:
        return roc_auc_score(y_real, y_pred[:, 0])
    else:
        return roc_auc_score(y_real, 1 - y_pred)


REGRESSION_METRICS = {
    "mse": mean_squared_error,
    "abs": mean_absolute_error
}

BINARY_CLASSIFICATION_METRICS = {
    "acc": inverse_accuracy,
    # "roc_auc": inverse_roc_auc
}


def get_regression_metric(metric: str) -> callable:
    if metric in MetricsConstants.REGRESSION_METRICS:
        return REGRESSION_METRICS[metric]


def get_binary_classification_metric(metric: str) -> callable:
    if metric in MetricsConstants.BINARY_METRICS:
        return BINARY_CLASSIFICATION_METRICS[metric]


__all__ = [
    'discriminability', 'certainty', 'divergency',
    'weighted_interpolated_error', "get_regression_metric"
]



import os
import pickle
from typing import Union
from time import time
from abc import ABC, abstractmethod

import joblib
import numpy as np
from pandas.core.frame import DataFrame

from bliz.builder.save_load_model import ModelFileHandler


class BaseModel(ABC):
    def __init__(
        self,
    ):
        self.model_name = str(int(time()))
        self.model = None

    @abstractmethod
    def fit(
        self,
        X,
        y
    ) -> None:
        pass

    @abstractmethod
    def predict(
        self,
        X,
    ) -> Union[DataFrame, np.ndarray, list]:
        pass

    def save(
        self,
        base_path,
    ):
        model_dir = os.path.join(base_path, self.model_name)
        if not os.path.exists(model_dir):
            os.makedirs(model_dir)
        model_path = os.path.join(
            model_dir, "{}.joblib".format(
                self.model_name))
        try:
            with open(model_path, 'wb') as model_file:
                joblib.dump(
                    self,
                    model_file,
                )
            return model_dir
        except pickle.PicklingError as e:
            print("Cannot picke model at {p} due to {e}".format(
                p=model_path, e=e))
            raise e

    def load(
        self,
        model_path,
    ):
        self.model = ModelFileHandler().load(model_path)



import logging

log_format = """{"time": "%(asctime)s","level": "%(levelname)s",
"log_name": "%(name)s", "function": "%(funcName)s","message": "%(message)s"}"""\
    .strip().replace('\n', '')


class Logger(object):
    def __init__(self, name, level="INFO"):
        self.logger = logging.getLogger(name)
        self.handler = logging.StreamHandler()
        self.handler.setFormatter(logging.Formatter(log_format))
        self.logger.addHandler(self.handler)
        self.logger.setLevel(level)

    def get_logger(self) -> logging.Logger:
        return self.logger



import pandas as pd
from sklearn.model_selection import train_test_split
from time import time


def prepare_data_for_training(
    df: pd.core.frame.DataFrame,
    target: str,
    index_column: str = None,
    validation_test_size: float = 0.2,
    verbose: bool = False,
):
    """takes input data and setting index, seperating target column
    and retuns a random split of given size

    Args:
        df(pd.core.frame.DataFrame): input data
        target(str): target column name
        index_column(str): index column name, default to None (no indexing)
        validation_test_size(float): test set size 0.0-1.0 if 0 all will be training set
            defaults to 0.2
        verbose(bool): prints out the shape of test and train datasets

    Returns:
        _x, test_x, _y, test_y: a tuple of datasets - train, test

    """
    if index_column:
        if index_column in df.columns:
            _df = df.set_index(index_column)
        else:
            raise KeyError("{} not in dataframe columns: [{}]".format(
                index_column,
                df.columns
            ))
    else:
        _df = df
    if target in df.columns:
        _target = _df[target]
        _data = _df.drop(target, axis=1)
    else:
        raise KeyError("{} not in dataframe columns: [{}]".format(
            target,
            df.columns
        ))
    if validation_test_size == 0:
        _x, test_x, _y, test_y = _data, pd.DataFrame(), _target, pd.Series()
    else:
        _x, test_x, _y, test_y = train_test_split(
            _data,
            _target,
            test_size=validation_test_size,
            random_state=int(time()),
        )
    if verbose:
        print("shape of training data = {}".format(_x.shape))
        print("shape of training data target = {}".format(_y.shape))
        print("shape of validation data = {}".format(test_x.shape))
        print(
            "shape of validation data target = {}\n".format(
                test_y.shape))
    return _x, _y.values, test_x, test_y.values



import json

import numpy as np

from bliz.evaluation.utils import parse_ndarray_as_float_list
from bliz.evaluation.experiment import Experiment
from bliz.evaluation.simulation_data import Metrics
from bliz.evaluation.metrics import discriminability, certainty
from bliz.evaluation.plotting.hist import single_hist
from bliz.evaluation.plotting.result_data import ResultData


class MonteCarloSimulation(object):
    """facilitates the experiments conducted, and calculating the metrics
    """

    def __init__(
            self,
            exp_type: str
    ):
        """initiates monte carlo simulation

        Args:
            W: weights for ratio between metrics
            bins: list of bin edges
        Raises:
            ValueError: if lists not of same length
        """
        self.experiments = []
        self.scores = {}
        self.metrics = None
        self.exp_type = exp_type

    def load_experiment(self,
                        real: list,
                        real_trained: np.ndarray,
                        model: list,
                        ):
        """loading a single experiment to simulation

        Args:
            real(list): list of ground truth results of the test set
            real_trained(list): list of ground truth results of the training set
            model(list): list of subjected-model predictions
            rand(int): scale of random samples, ir R in (0, n), defaults to 1
            others(dict): dictionary of other models predictions.
        """
        self.experiments.append(
            Experiment(self.exp_type, real, real_trained, model))

    def digest(self, metric):
        """calculates the full simulation results on the experiments
        loaded thus far

        Args:
            metric(callable): the metric to calculate results on (array-like, array-like) -> float
                              defaults to sklearn.metrics.mean_squared_error
        """
        _scores = []
        for exp in self.experiments:
            _scores.append(exp.score(metric))
        self.scores["model"] = []
        self.scores["rand"] = []
        self.scores["mean"] = []
        for s in _scores:
            self.scores["model"].append(s.Model)
            self.scores["rand"].append(s.Rand)
            self.scores["mean"].append(s.Mean)
        self.metrics = Metrics(
            discriminability(self.scores["model"], self.scores["mean"],
                             self.scores["rand"]),
            certainty(self.scores["model"], self.scores["rand"]),
        )

    def get_metrics(self):
        """returns the Metrics namedTuple

        Returns:
            metrics (Metrics)
        """
        return self.metrics

    def metrics_as_dict(self):
        """returns the Metrics as dict

        Returns:
            metrics (dict): dictionary of metrics
        """
        if self.metrics:
            return {
                "discriminability": self.metrics.Discriminability,
                "certainty": self.metrics.Certainty,
            }
        else:
            return None

    def metrics_to_json(
            self,
            path: str,
    ):
        """saves MC result metrics as .json file

        Args:
            path(str): path to save json file
            filename(str): filename to be used in saving
        """
        with open(path, 'w+') as output_file:
            output_file.write(json.dumps(self.metrics_as_dict()))

    def save_experiment_summery(
            self,
            path: str,
    ):
        """saves a summery report of the experiments

        Args:
            path(str): path to save summry report json file to
        """
        experiment_summery = {}
        for i, exp in enumerate(self.experiments):
            experiment_id = "experiment_{}".format(i)
            experiment_summery[experiment_id] = {
                "real": parse_ndarray_as_float_list(exp.experiment_data.Real),
                "model": parse_ndarray_as_float_list(exp.experiment_data.Model),
                "mean": parse_ndarray_as_float_list(exp.experiment_data.Mean),
                "rand": parse_ndarray_as_float_list(exp.experiment_data.Rand),
            }
        with open(path, 'w+') as output_file:
            output_file.write(json.dumps(experiment_summery))

    def plot(self, path=None, title=None):
        """plots simulation histograms

        Args:
            path(str): path to save plots to
        """
        try:
            max_scores = [max(v) for k, v in self.scores.items()]
            bins = np.linspace(0, min(1.0, max(max_scores)),
                               max(int(len(self.scores["model"]) / 10), 100))
            plots = [ResultData(k, v, None) for k, v in self.scores.items()]
            if path:
                single_hist(data=plots, bins=bins, path=path, title=title)
            else:
                return single_hist(data=plots, bins=bins, title=title)
        except Exception as e:
            raise e



import numpy as np


def min_max_norm(y):
    """calculates the min max norm of an array

    Args:
        y(array-like): an array of int or floats

    Returns:
        out(array-like): a normalized array

    Raises:
        ValueError: if y is not array-like
    """
    try:
        _y = np.asarray(y)
    except BaseException as be:
        raise ValueError(
            "y must be scalar or array like :: {}".format(
                str(be)))
    return (_y - _y.min()) / (_y.max() - _y.min())


NORM_MAP = {
    "min_max": min_max_norm
}


def get_normalizer(normalizer: str):
    """returns a normalizer function using a string

    Args:
        normalizer(str): normalizer identifier key

    Returns:
        normalizer(Callable): a normalizing method

    Raises:
        KeyError: if normalizer is not available or key not found
    """
    if normalizer in NORM_MAP:
        return NORM_MAP[normalizer]
    else:
        raise KeyError("normalizer is not in available normalizers: [{}]".format(
            NORM_MAP.keys()
        ))



import os
from typing import Union
import random

import pandas as pd
from tqdm import tqdm
import matplotlib.pyplot as plt

from bliz.builder.data_prep import prepare_data_for_training
from bliz.builder.save_load_model import load_model
from bliz.builder.utils import is_regression_metric, is_binary_classification_metric
from bliz.evaluation import MonteCarloSimulation
from bliz.evaluation.metrics import get_regression_metric, get_binary_classification_metric
from bliz.evaluation.norms import get_normalizer
from bliz.logger import Logger

BUILD_DIR_NAME = "build"


class Builder(object):

    logger = Logger("Builder").get_logger()

    def __init__(
        self,
        path_to_model: str,
        exp_type: str,
        data: pd.core.frame.DataFrame,
        target: str,
        metric: callable,
        index: str = None
    ):
        """Base builder class, handles the MC iteration

        Args:
            data(pd.DataFrame):
            path_to_model(str): path to a folder containing model.py file
            target(str): name of column in data to target for
            index(str): name of colummn in data to use as index, default None
        """
        self.data = data
        self.path_to_model = path_to_model
        self.target = target
        self.index_column = index
        self.final_model = None
        self.result = None
        self.MC_simulation = MonteCarloSimulation(exp_type)

    def _build(
        self,
        n_experiments: int,
        metric,
    ):
        """Build and evaluates the model loaded

        Args:
            n_experiments(int): number of experiment to execute
        """
        self.logger.info("Initiating {} Epochs".format(n_experiments))
        Model = load_model(self.path_to_model)
        with tqdm(total=n_experiments, desc=" Training Model") as bar:
            for _ in range(n_experiments):
                X, y, v_X, v_y = prepare_data_for_training(
                    df=self.data,
                    target=self.target,
                    index_column=self.index_column,
                    validation_test_size=random.uniform(0.15, 0.25),
                )
                this_model = Model()
                this_model.fit(X, y)
                this_prediction = this_model.predict(v_X)
                self.MC_simulation.load_experiment(
                    v_y,
                    y,
                    this_prediction
                )
                bar.update(1)
        self.MC_simulation.digest(metric=metric)
        self.result = self.MC_simulation.metrics_as_dict()

    def finalize_model(self,):
        """trains the model on the entire dataset

        """
        X, y, _, _ = prepare_data_for_training(
            df=self.data,
            target=self.target,
            index_column=self.index_column,
            validation_test_size=0,
        )
        self.logger.info("Finalzing model")
        Model = load_model(self.path_to_model)
        self.logger.info("Training model on all data")
        final_model = Model()
        final_model.fit(X, y)
        final_model.save("playground")
        self.final_model = Model()
        self.final_model.fit(X, y)

    def get_results(self) -> dict:
        """returns the building stage results

        Returns:
            results(dict): a dictionary with the model building results

        """
        if self.result:
            return self.result
        else:
            raise RuntimeError("You must use build() to get results")

    def plot(self, title=None):
        """plots the simulation histogram summery to screen

        Args:
            title(str): plot's title

        """
        plt.clf()
        self.MC_simulation.plot(title=title)
        plt.show()

    def save(self, plot=True, summery=False, data=False):
        """Saves the model to the model directory

        Args:
            plot(bool): saves the simulation histogram as png default True
            summery(bool): saves the summry of all experiments ran default False
            data(bool): saves the data used in traininig default False
        """
        if self.final_model:
            print("saving model")
            print(type(self.final_model))
            model_dir = self.final_model.save(
                os.path.join(self.path_to_model, BUILD_DIR_NAME))
            self.logger.info("saved model to {}".format(model_dir))
        else:
            raise RuntimeError(
                "You did not finalize model thus no model will be saved, use .finalize_model() method to save model")
        if self.result:
            self.logger.info("saving model metrics")
            self.MC_simulation.metrics_to_json(os.path.join(
                model_dir, "{}_metadata.json".format(self.final_model.model_name)))
            if plot:
                self.logger.info("saving simultion plot")
                self.MC_simulation.plot(path=model_dir)
            else:
                self.logger.info("plot=False will not save evaluation plot")
        else:
            raise RuntimeError("You must use build() to save")
        if summery:
            self.logger.info("saving experiment summery")
            self.MC_simulation.save_experiment_summery(os.path.join(
                model_dir, "{}_summery.json".format(self.final_model.model_name)))
        else:
            self.logger.info(
                "summery = False, will not save experiment summery")
        if data:
            self.logger.info("saving input data")
            self.data.to_csv(os.path.join(
                model_dir, "{}_data.csv".format(self.final_model.model_name)))
        else:
            self.logger.info("data = False, will not save experiment data")
        return model_dir


class RegressionBuilder(Builder):
    def __init__(
        self,
        path_to_model: str,
        data: pd.core.frame.DataFrame,
        target: str,
        metric: Union[str, callable],
        normalizer: Union[str, callable] = None,
        index: str = None
    ):
        if normalizer:
            self.logger.info(
                "normalizing data target, this will duplicate data space in mem")
            norm_data = data.copy()
            if hasattr(normalizer, "__call__"):
                norm_data[target] = normalizer(norm_data[target])
            elif isinstance(normalizer, str):
                norm_data[target] = get_normalizer(
                    normalizer)(norm_data[target])
            else:
                raise ValueError("normalizer should be either str or callable")
            super().__init__(path_to_model, "reg", norm_data, target, metric)
        else:
            super().__init__(path_to_model, "reg", data, target, metric)
        if hasattr(metric, "__call__"):
            assert is_regression_metric(metric)
            self._metric = metric
        elif isinstance(metric, str):
            self._metric = get_regression_metric(metric)
        else:
            raise ValueError(
                "metric should be str or callable, not {}".format(
                    type(metric)))

    def build(
        self,
        n_experiments: int,
    ):
        self._build(n_experiments, self._metric)


class BinaryClassificationBuilder(Builder):
    def __init__(
        self,
        path_to_model: str,
        data: pd.core.frame.DataFrame,
        target: str,
        metric: Union[str, callable],
        index: str = None
    ):
        super().__init__(path_to_model, "binary", data, target, metric)
        if hasattr(metric, "__call__"):
            assert is_binary_classification_metric(metric)
            self._metric = metric
        elif isinstance(metric, str):
            self._metric = get_binary_classification_metric(metric)
        else:
            raise ValueError(
                "metric should be str or callable, not {}".format(
                    type(metric)))

    def build(
        self,
        n_experiments: int,
    ):
        self._build(n_experiments, self._metric)



import os
import sys
import joblib
import pickle
import importlib


class ModelFileHandler(object):
    def __init__(
        self,
        model=None,
    ):
        self.model = model
        self.path = None

    def save(
        self,
        path=None,
    ):
        model_name = self.model.model_name
        model_dir = os.path.join(path, model_name)
        if not os.path.exists(model_dir):
            os.makedirs(model_dir)
        model_path = os.path.join(model_dir, "{}.joblib".format(model_name))
        try:
            out = self.model
            print("trying to pickle {} of type {}".format(out, type(out)))
            with open(model_path, 'wb') as model_file:
                joblib.dump(
                    out,
                    model_file,
                )
            return model_dir
        except pickle.PicklingError as e:
            print("Cannot picke model at {p} due to {e}".format(
                p=model_path, e=e))
            return model_dir

    def load(
        self,
        model_dir: str,
    ):
        model_name = "{}.joblib".format(os.path.split(model_dir)[-1])
        model_path = os.path.join(model_dir, model_name)
        if not os.path.exists(model_path):
            raise FileExistsError("file {} is not found in folder {} ".format(
                model_name, model_dir
            ) + "cannot load")
        with open(model_path, 'rb') as model_file:
            return joblib.load(model_file)


def load_model(
    model_source: str,
):
    # LOAD MODEL
    spec = importlib.util.spec_from_file_location(
        name='model',
        location=os.path.join(model_source, "model.py")
    )
    model = importlib.util.module_from_spec(
        spec
    )
    spec.loader.exec_module(model)
    sys.modules['model'] = model
    return model.Model



import numpy as np

from collections import namedtuple
from sklearn.metrics import classification_report, confusion_matrix


Doc = namedtuple(
    'Doc',
    [
        "version",
        "type",
        "algo",
        "param",
        "cv",
    ]
)


def compress_regression_results(l, true_condition=lambda x: x >= 0.6):
    out = []
    for x in list(l):
        if true_condition(x):
            out.append(1)
        else:
            out.append(0)
    return out


def generate_model_documentation(
    model_version: str,
    model_type: str,
    model_algorithm: str,
    model_parameter_tuning: str = None,
    model_cv: str = None,
):
    return (
        "\n---------- Model Details:\n\n" +
        "Model Version == {}\n".format(model_version) +
        "Model Type == {}\n".format(model_type) +
        "Model Algorithm == {}\n".format(model_algorithm) +
        "Model Parameter Tuning == {}\n".format(model_parameter_tuning) +
        "Model CV == {}\n".format(model_cv)
    )


def create_classification_report(y_real, y_pred):
    p = compress_regression_results(list(y_pred))
    r = compress_regression_results(list(y_real))

    for y_p, y_r_p, y_r, y_r_r in zip(
        p,
        list(y_pred),
        r,
        list(y_real),
    ):
        print("Predicted {rp} ~ {p} for result {rr} ~ {r}".format(
            p=y_p,
            rp=y_r_p,
            r=y_r,
            rr=y_r_r,
        ))
    print("\n{}".format(
        classification_report(
            r,
            p,
            target_names=["bottom_tier", "top_tier"],
        )
    )
    )
    tn, fp, fn, tp = confusion_matrix(p, r).ravel()
    print(
        "tn = {} \n".format(tn / len(p)) +
        "tp = {} \n".format(tp / len(p)) +
        "fn = {} \n".format(fn / len(p)) +
        "fp = {} \n".format(fp / len(p))
    )
    print(
        "Precision = {}\n".format(round(tp / (tp + fp), 2)) +
        "Recall = {}\n".format(round(tp / (tp + fn), 2))
    )


def min_max_norm(y: np.ndarray) -> np.ndarray:
    return (y - y.min()) / (y.max() - y.min())


def is_regression_metric(metric: callable) -> bool:
    real = np.asarray([0.1, 0.33, 0.44])
    pred_close = np.asarray([0.11, 0.34, 0.45])
    pred_far = np.asarray([0.3, 0.6, 0.9])
    if (not metric(real, real) == 0.0 and
            not metric(real, real) < metric(real, pred_close) < metric(real, pred_far)):
        return False
    else:
        return True


def is_binary_classification_metric(metric: callable) -> bool:
    real = np.asarray([1, 1, 0])
    pred_close = np.asarray([1, 0, 0])
    pred_far = np.asarray([0, 0, 1])
    if (not metric(real, real) == 0.0 and
            not metric(real, real) < metric(real, pred_close) < metric(real, pred_far)):
        return False
    else:
        return True



from .builder.builder import RegressionBuilder, BinaryClassificationBuilder
from .builder.save_load_model import ModelFileHandler
from .evaluation.metrics import weighted_interpolated_error


__all__ = [
    "RegressionBuilder",
    "ModelFileHandler",
    "BinaryClassificationBuilder",
    "weighted_interpolated_error"
]



import random
from time import time
from collections import namedtuple

import numpy as np

from bliz.evaluation.simulation_data import ExperimentData, Score
from bliz.evaluation.utils import validate_multiple_lists_length
from bliz.evaluation.metrics import WD
from bliz.evaluation.constants import ExperimentConstants

random.seed(int(time()) % 10**4)


def length_error(data_length):
    raise ValueError(
        "all data should be of the same length - {}".format(data_length))


def _divergence_by_wd(dist, ref):
    """calculated the divergence between two distributions, meaning:
    the inverse of the distance between them, the more they are evenly
    distributed == divergence is higher

    Args:
        dist(array-like): original distribution
        ref(array-like): refrence distribution

    Returns:
        abs_wd
    """
    abs_wd = 1 / abs(WD(dist, ref))
    if abs_wd > 0.0:
        return abs_wd
    else:
        return np.inf


def reg_mean(y, size):
    return np.full(size, np.asarray(y).mean())


def binary_mean(y, size):
    return np.full(size, np.bincount(y).argmax())


def choice_rand(y, size):
    return np.random.choice(y, size=size)


BASE_DIST = {
    "reg": {
        "mean": reg_mean,
        "rand": choice_rand
    },
    "binary": {
        "mean": binary_mean,
        "rand": choice_rand
    }
}


class Experiment(object):
    """
    gets monte carlo exp results and returns metrics
    """

    ExperimentData = namedtuple('ExperimentData', [
        "Real",
        'Model',
        'Rand',
        'Mean',
        # 'OtherModels',
    ])

    Score = namedtuple('Score', [
        'Model',
        'Rand',
        'Mean',
    ])

    def __init__(
            self,
            exp_type: str,
            real: list,
            real_trained: np.ndarray,
            model: list,
    ):
        self.experiment_data = None
        if exp_type in ExperimentConstants.TYPES:
            self.exp_type = exp_type
        else:
            raise ValueError("exp_type must be one of: [{}]".format(
                ExperimentConstants.TYPES))
        self._load(
            real_results=real,
            real_trained_results=real_trained,
            model_prediction=model,
        )
        self.experiment_results = None

    def _load(
        self,
        real_results: list,
        real_trained_results: np.ndarray,
        model_prediction: list,
    ):
        """loads experiment data into ExperimentData object

        Args:
            real_results(list): real results of this experiments
            real_trained_results(np.ndarray): target vector of the trained data in this experiment
            model_prediction(list): predictions of the model
            random_scale(int): what is the scale [0, X] from which random results would be selected
            other_models_predictions(dict): a dictionary of other models model_name->list_of_socres

        Raises:
            ValueError: if there is a mismatch in length of any of the arguments
        """
        if validate_multiple_lists_length(
            real_results,
            model_prediction,
        ):
            size = len(real_results)
            random_data = BASE_DIST[self.exp_type]["rand"](
                real_trained_results, size)
            mean_data = BASE_DIST[self.exp_type]["mean"](
                real_trained_results, size)
            self.experiment_data = ExperimentData(
                real_results, model_prediction, random_data,
                mean_data)
        else:
            raise length_error(len(real_results))

    def score(self, metric) -> Score:
        """calculates the score of this model based on the metrics

        Args:
            metric(callable): which metric should calcultae the error bw 2 results sets

        Returns:
            score(Score): score object with all the metric calculated scores
        """
        self.experiment_results = Score(
            metric(self.experiment_data.Real, self.experiment_data.Model),
            metric(self.experiment_data.Real, self.experiment_data.Rand),
            metric(self.experiment_data.Real, self.experiment_data.Mean),
        )
        return self.experiment_results



import numpy as np


def validate_multiple_lists_length(*lists) -> bool:
    """Validates that a list of lists is of the same length

    Args:
        lists(list): a list of lists to be checked

    Retuens:
        b(bool): True if all list of the same length, False else. 

    """
    list_len = -1
    for l in lists:
        try:
            iter(l)
        except BaseException:
            return False
        if list_len == -1:  # first list
            list_len = len(l)
        else:
            if list_len != len(l):
                return False
    return True


def close_enough(a, b, precision=3) -> bool:
    """checks weather to Array-like are close enough to each other

    Args:
        a: first argument
        b: second argument
        precision(int): the decimal level of accuracy, 3 == 0.001

    Returns:
        out(bool): True, if a and b are close enough on precision, False else.

    """

    def _check_precision(x, y):
        if round(x, precision) == round(y, precision):
            return True
        else:
            return False

    try:
        if a == b:
            return True
    except ValueError:  # comparing numpy.ndarray for example
        pass
    try:
        iter(a)
        iter(b)
    except BaseException:  # objects are not iterable == scalar
        return _check_precision(a, b)
    out = True
    for i, j in zip(
            a,
            b):  # if we got here we need to iterate over all args in the array
        out = out and _check_precision(i, j)
    return out


def _sample_to_bin(a, bins):
    """return the bin number of each sample

    Args:
        a: list of array like, same length
        bins: list of bin edges
    Retrurns:
        l: list same length as 'a' containing the bin number instead of values
    """
    bins_list = []
    for x in a:
        x_bin, _ = np.histogram(x, bins)
        bins_list.append(x_bin.argmax())
    return bins_list


def samples_to_bin_numbers(*lists, bins):
    """return the bin number of each sample

    Args:
        lists: list of array like, same length
        bins: list of bin edges
    Raises:
        ValueError: if lists not of same length
    Retruns:
        lists_out: list same length as *lists containing the bin number instead of values
    """
    if validate_multiple_lists_length(lists):
        lists_out = []
        for l in lists:
            lists_out.append(_sample_to_bin(l, bins))
        return tuple(lists_out)
    else:
        raise ValueError("lists should all have the same length!")


def parse_ndarray_as_float_list(arr: np.ndarray) -> list:
    """returns a list of python floats rather than numpy.ndarray of float32

    Args:
        arr(numpy.ndarray): an array of float32

    Returns:
        lst(list): list of float
    """
    return arr.tolist()


def is_binary(a) -> bool:
    if len(np.asarray(a).shape) >= 1:
        return False
    if (np.bincount(a) == 2
        and (np.max(a) == 1 or np.max(a) == 0)
            and (np.min(a) == 0 or np.min(a) == 1)):
        return True
    else:
        return False


__all__ = [
    "samples_to_bin_numbers", "_sample_to_bin", "close_enough",
    "validate_multiple_lists_length"
]



from collections import namedtuple

ExperimentData = namedtuple('ExperimentData', [
    "Real",
    'Model',
    'Rand',
    'Mean',
])

Score = namedtuple('Score', [
    'Model',
    'Rand',
    'Mean',
])

Metrics = namedtuple('Metrics',
                     [
                         'Discriminability',
                         'Certainty',
                     ]
                     )



from setuptools import setup

setup(
    # Needed to silence warnings (and to be a worthwhile package)
    name='bliz',
    url='https://github.com/miararoy/bliz',
    author='Roy Miara',
    author_email='miararoy@gmail.com',
    # Needed to actually package something
    packages=[
        "bliz.builder",
        "bliz.evaluation"
    ],
    # Needed for dependencies
    install_requires=[
        'numpy==1.15.2',
        'pandas==0.25.0',
        'scipy==1.3.0',
        'matplotlib==3.0.2',
        'seaborn==0.9.0',
        'scikit-learn==0.21.2',
        'joblib==0.13.2',
        'tqdm==4.31.1',
    ],
    # *strongly* suggested for sharing
    version='0.1.2',
    description='Utilities for dataframes',
    # We will also need a readme eventually (there will be a warning)
    long_description=open('README.md').read(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6'
)
